@@ -7,6 +7,16 @@
 #endif
 #include "_ph_postprocess.h"
 
+/* =========== Free-threading support ======== */
+
+#ifdef Py_GIL_DISABLED
+#define MUTEX_LOCK(m) PyMutex_Lock(m)
+#define MUTEX_UNLOCK(m) PyMutex_Unlock(m)
+#else
+#define MUTEX_LOCK(m)
+#define MUTEX_UNLOCK(m)
+#endif
+
 /* =========== Common stuff ======== */
 
 #define MAX_ENCODERS 20
@@ -99,6 +109,9 @@ typedef struct {
     uint8_t *data;                              // pointer to data after decoding
     int stride;                                 // time when it get filled depends on `remove_stride` value
     PyObject *file_bytes;                       // private
+#ifdef Py_GIL_DISABLED
+    PyMutex decode_mutex;                       // protects lazy decode in free-threaded builds
+#endif
 } CtxImageObject;
 
 static PyTypeObject CtxImage_Type;
@@ -142,7 +155,7 @@ static PyObject* _CtxWriteImage_add_plane(CtxWriteImageObject* self, PyObject* a
         real_stride = real_stride * 2;
     if (stride_in == 0)
         stride_in = real_stride;
-    if (stride_in * height > buffer.len) {
+    if ((Py_ssize_t)stride_in * height > buffer.len) {
         PyBuffer_Release(&buffer);
         PyErr_SetString(PyExc_ValueError, "image plane does not contain enough data");
         return NULL;
@@ -328,7 +341,7 @@ static PyObject* _CtxWriteImage_add_plane_la(CtxWriteImageObject* self, PyObject
         real_stride = real_stride * 2;
     if (stride_in == 0)
         stride_in = real_stride;
-    if (stride_in * height > buffer.len) {
+    if ((Py_ssize_t)stride_in * height > buffer.len) {
         PyBuffer_Release(&buffer);
         PyErr_SetString(PyExc_ValueError, "image plane does not contain enough data");
         return NULL;
@@ -433,7 +446,7 @@ static PyObject* _CtxWriteImage_add_plane_l(CtxWriteImageObject* self, PyObject*
         real_stride = real_stride * 2;
     if (stride_in == 0)
         stride_in = real_stride;
-    if (stride_in * height > buffer.len) {
+    if ((Py_ssize_t)stride_in * height > buffer.len) {
         PyBuffer_Release(&buffer);
         PyErr_SetString(PyExc_ValueError, "image plane does not contain enough data");
         return NULL;
@@ -524,6 +537,14 @@ static PyObject* _CtxWriteImage_set_nclx_profile(CtxWriteImageObject* self, PyOb
     Py_RETURN_NONE;
 }
 
+static PyObject* _CtxWriteImage_set_pixel_aspect_ratio(CtxWriteImageObject* self, PyObject* args) {
+    unsigned int aspect_h, aspect_v;
+    if (!PyArg_ParseTuple(args, "II", &aspect_h, &aspect_v))
+        return NULL;
+    heif_image_set_pixel_aspect_ratio(self->image, aspect_h, aspect_v);
+    Py_RETURN_NONE;
+}
+
 static PyObject* _CtxWriteImage_encode(CtxWriteImageObject* self, PyObject* args) {
     /* ctx: CtxWriteObject, primary: int */
     CtxWriteObject* ctx_write;
@@ -657,6 +678,7 @@ static struct PyMethodDef _CtxWriteImage_methods[] = {
     {"add_plane_la", (PyCFunction)_CtxWriteImage_add_plane_la, METH_VARARGS},
     {"set_icc_profile", (PyCFunction)_CtxWriteImage_set_icc_profile, METH_VARARGS},
     {"set_nclx_profile", (PyCFunction)_CtxWriteImage_set_nclx_profile, METH_VARARGS},
+    {"set_pixel_aspect_ratio", (PyCFunction)_CtxWriteImage_set_pixel_aspect_ratio, METH_VARARGS},
     {"encode", (PyCFunction)_CtxWriteImage_encode, METH_VARARGS},
     {"set_exif", (PyCFunction)_CtxWriteImage_set_exif, METH_VARARGS},
     {"set_xmp", (PyCFunction)_CtxWriteImage_set_xmp, METH_VARARGS},
@@ -821,6 +843,9 @@ PyObject* _CtxAuxImage(struct heif_image_handle* main_handle, heif_item_id aux_i
     ctx_image->file_bytes = file_bytes;
     ctx_image->stride = get_stride(ctx_image);
     strcpy(ctx_image->decoder_id, decoder_id);
+#ifdef Py_GIL_DISABLED
+    ctx_image->decode_mutex = (PyMutex){0};
+#endif
     Py_INCREF(file_bytes);
     return (PyObject*)ctx_image;
 }
@@ -872,6 +897,9 @@ PyObject* _CtxDepthImage(struct heif_image_handle* main_handle, heif_item_id dep
     ctx_image->file_bytes = file_bytes;
     ctx_image->stride = get_stride(ctx_image);
     strcpy(ctx_image->decoder_id, decoder_id);
+#ifdef Py_GIL_DISABLED
+    ctx_image->decode_mutex = (PyMutex){0};
+#endif
     Py_INCREF(file_bytes);
     return (PyObject*)ctx_image;
 }
@@ -955,6 +983,9 @@ PyObject* _CtxImage(struct heif_image_handle* handle, int hdr_to_8bit,
     ctx_image->file_bytes = file_bytes;
     ctx_image->stride = get_stride(ctx_image);
     strcpy(ctx_image->decoder_id, decoder_id);
+#ifdef Py_GIL_DISABLED
+    ctx_image->decode_mutex = (PyMutex){0};
+#endif
     Py_INCREF(file_bytes);
     return (PyObject*)ctx_image;
 }
@@ -1257,16 +1288,26 @@ int decode_image(CtxImageObject* self) {
 }
 
 static PyObject* _CtxImage_stride(CtxImageObject* self, void* closure) {
-    if (!self->data)
-        if (!decode_image(self))
+    MUTEX_LOCK(&self->decode_mutex);
+    if (!self->data) {
+        if (!decode_image(self)) {
+            MUTEX_UNLOCK(&self->decode_mutex);
             return NULL;
+        }
+    }
+    MUTEX_UNLOCK(&self->decode_mutex);
     return PyLong_FromSsize_t(self->stride);
 }
 
 static PyObject* _CtxImage_data(CtxImageObject* self, void* closure) {
-    if (!self->data)
-        if (!decode_image(self))
+    MUTEX_LOCK(&self->decode_mutex);
+    if (!self->data) {
+        if (!decode_image(self)) {
+            MUTEX_UNLOCK(&self->decode_mutex);
             return NULL;
+        }
+    }
+    MUTEX_UNLOCK(&self->decode_mutex);
     return PyMemoryView_FromMemory((char*)self->data, self->stride * self->height, PyBUF_READ);
 }
 
@@ -1352,6 +1393,17 @@ static PyObject* _CtxImage_get_aux_type(CtxImageObject* self, PyObject* arg_imag
     return aux_type;
 }
 
+static PyObject* _CtxImage_pixel_aspect_ratio(CtxImageObject* self, void* closure) {
+    #if LIBHEIF_HAVE_VERSION(1,19,0)
+        uint32_t aspect_h, aspect_v;
+        int has_pasp = heif_image_handle_get_pixel_aspect_ratio(self->handle, &aspect_h, &aspect_v);
+        if (has_pasp) {
+            return Py_BuildValue("(II)", aspect_h, aspect_v);
+        }
+    #endif
+    Py_RETURN_NONE;
+}
+
 /* =========== CtxImage Experimental Part ======== */
 
 static PyObject* _CtxImage_camera_intrinsic_matrix(CtxImageObject* self, void* closure) {
@@ -1415,6 +1467,7 @@ static struct PyGetSetDef _CtxImage_getseters[] = {
     {"data", (getter)_CtxImage_data, NULL, NULL, NULL},
     {"depth_image_list", (getter)_CtxImage_depth_image_list, NULL, NULL, NULL},
     {"aux_image_ids", (getter)_CtxImage_aux_image_ids, NULL, NULL, NULL},
+    {"pixel_aspect_ratio", (getter)_CtxImage_pixel_aspect_ratio, NULL, NULL, NULL},
     {"camera_intrinsic_matrix", (getter)_CtxImage_camera_intrinsic_matrix, NULL, NULL, NULL},
     {"camera_extrinsic_matrix_rot", (getter)_CtxImage_camera_extrinsic_matrix_rot, NULL, NULL, NULL},
     {NULL, NULL, NULL, NULL, NULL}
@@ -1698,17 +1751,21 @@ static int setup_module(PyObject* m) {
     return 0;
 }
 
+static PyModuleDef_Slot module_slots[] = {
+    {Py_mod_exec, setup_module},
+#ifdef Py_GIL_DISABLED
+    {Py_mod_gil, Py_MOD_GIL_NOT_USED},
+#endif
+    {0, NULL}
+};
+
 PyMODINIT_FUNC PyInit__pillow_heif(void) {
     static PyModuleDef module_def = {
         PyModuleDef_HEAD_INIT,
-        "_pillow_heif", /* m_name */
-        NULL,           /* m_doc */
-        -1,             /* m_size */
-        heifMethods,    /* m_methods */
+        .m_name = "_pillow_heif",
+        .m_methods = heifMethods,
+        .m_slots = module_slots,
     };
 
-    PyObject* m = PyModule_Create(&module_def);
-    if (setup_module(m) < 0)
-        return NULL;
-    return m;
+    return PyModuleDef_Init(&module_def);
 }
